package com.yuandian.dataflow.statemachine.grpc;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.time.Duration;
import java.time.Instant;
import java.util.HashMap;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import org.reflections.Reflections;

import com.yuandian.dataflow.proto.Processor.PacketsProcessorRequest;
import com.yuandian.dataflow.proto.Processor.ProcessorResponse;
import com.yuandian.dataflow.proto.ProcessorServerGrpc.ProcessorServerImplBase;
import com.yuandian.dataflow.statemachine.annotations.GrpcProcessor;
import com.yuandian.dataflow.statemachine.annotations.MasterRegister;
import com.yuandian.dataflow.statemachine.master.MasterExecute;
import com.yuandian.dataflow.statemachine_old.MasterFactory;
import com.yuandian.dataflow.utils.Utils;

import io.grpc.BindableService;
import io.grpc.Server;
import io.grpc.ServerBuilder;
import io.grpc.ServerServiceDefinition;
import io.grpc.stub.StreamObserver;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

@Slf4j
@Getter
@Setter
public class ProcessorServer {

    public static void main(String[] args) throws Exception {
        // ServerBuilder builder = ServerBuilder.forPort(0);
        // var server = builder.build().start();
        // log.info("{}", server.getPort());
        var server =  new ProcessorServer();
        server.grpcServer.awaitTermination();
    }

    private Server grpcServer;
 

    public ProcessorServer()   {
        ServerBuilder builder = ServerBuilder.forPort(0);
 
        // 扫描注解RaftProccessor 注册
        var now = Instant.now();
        HashMap<String, Class<?>> scansMap = new HashMap<>();
        var traces = Thread.currentThread().getStackTrace();
        var clsName = traces[traces.length - 1].getClassName();
        var packName = clsName.substring(0, Utils.indexOfAsPossible(clsName, ".", 3));
        log.info("获取 {} -> {} 下包的所有注解", clsName, packName);

        var refl = new Reflections(packName);
        Set<Class<?>> scans = refl.getTypesAnnotatedWith(GrpcProcessor.class);

        scans.forEach((pRaftClass) -> {
            scansMap.put(pRaftClass.getName(), pRaftClass);
        });
        log.info("扫描注解的时间(scan annotations cost time): {} ms", Duration.between(now, Instant.now()).toMillis());
        scansMap.forEach((name, pRaftClass) -> {
            try {
                builder.addService( (BindableService)pRaftClass.getDeclaredConstructor().newInstance() );
            } catch (InstantiationException | IllegalAccessException | IllegalArgumentException
                    | InvocationTargetException | NoSuchMethodException | SecurityException e) {
                log.info("{}",e.toString());
            }
        });

        refl.getTypesAnnotatedWith(MasterRegister.class).forEach((pClass) -> {
            try {
                MasterExecute execute = (MasterExecute) pClass.getDeclaredConstructor().newInstance();
                MasterFactory.registerMasterLoop(execute);
            } catch (InstantiationException | IllegalAccessException | IllegalArgumentException
                    | InvocationTargetException | NoSuchMethodException | SecurityException e) {
                log.info("{}", e.toString());

            }
        });


        grpcServer = builder.build();
        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                // Use stderr here since the logger may have been reset by its JVM shutdown hook.
                log.info("*** shutting down gRPC server since JVM is shutting down");
                try {
                    grpcServer.shutdown().awaitTermination(30, TimeUnit.SECONDS);
                } catch (InterruptedException e) {
                    log.error("{}", e.toString());
                }
                System.err.println("*** server shut down");
            }
        });
    }
 
}
